# core-image-builder
# Copyright (C) 2015 Canonical
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Business logic or the service lives here."""

import logging
import os
import subprocess
import tempfile

from uservice_utils.queue import MessageActions

from core_image_builder.constants import LOGGING_EXTRA
from core_image_builder.cloud import get_glance_client
from core_image_builder.utils import check_call


logger = logging.getLogger(__name__)


class ImageBuilderWorker(object):

    """A worker callable that contains all our main logic."""

    def __init__(self, config, builder):
        self._config = config
        self.publish_results = builder

    def __call__(self, payload):
        extra = LOGGING_EXTRA.copy()
        extra.update(payload)
        try:
            image_name = payload['image_name']
            channel = payload['channel']
            device = payload['device']
        except KeyError as e:
            logger.error(
                "Unable to deserialize message payload - "
                "rejecting message: %s",
                e,
                extra=extra
            )
            return MessageActions.Retry

        with tempfile.TemporaryDirectory() as tmpdir:
            logger.info("Beginning image download.", extra=extra)
            try:
                image_path = download_image(
                    image_name,
                    channel,
                    device,
                    tmpdir
                )
            except subprocess.CalledProcessError as e:
                logger.error(
                    "Unable to download core image: %s",
                    e,
                    extra=extra
                )
                return MessageActions.Retry
            logger.info("Ubuntu Core image downloaded OK.", extra=extra)

            try:
                nova_image_path = convert_nova_image(image_path)
            except subprocess.CalledProcessError as e:
                logger.error(
                    "Unable to convert core image to qcow2 image: %s",
                    e,
                    extra=extra)
                return MessageActions.Retry
            logger.info("Image converted to qcow2 OK.", extra=extra)

            try:
                logger.info("Beginning image upload to glance...",
                            extra=extra)
                glance_image_id = upload_image_to_glance(
                    nova_image_path,
                    self._config
                )
            except Exception as e:
                logger.error(
                    "Unable to upload image to glance: %s",
                    e,
                    extra=extra
                )
                return MessageActions.Retry
            logger.info("Image uploaded to glance OK.", extra=extra)

        payload['nova_image_id'] = glance_image_id
        self.publish_results(payload)
        logger.info("Processing completed.", extra=extra)
        return MessageActions.Acknowledge


def download_image(name, channel, device, tmpdir):
    """Download the ubuntu code image, return a path to it on disk."""
    image_path = os.path.join(tmpdir, 'core-{}.img'.format(name))
    udf_env = {
        # u-d-f fills '~/.cache/ubuntuimages/' on downloads,
        # not particulary useful to us, unless we only have one
        # worker and perform testing retries.
        'HOME': tmpdir,
    }
    # The folling uses hardcoded parameters for '--channel' and 'release'.
    # It is ignoring the channel value passed in.
    cmd = ['sudo',
           'ubuntu-device-flash',
           '--revision', name,
           'core',
           '--device', device,
           '--channel', 'edge',
           '--size', '3',
           '-o', image_path,
           '--developer-mode',
           '--cloud',
           'rolling']
    check_call(cmd, env=udf_env)
    cmd = ['sudo',
           'rm',
           '-rf',
           os.path.join(tmpdir, '.gnupg')]
    check_call(cmd, env=udf_env)
    return image_path


def convert_nova_image(image_path):
    """
    Convert a core image to a nova image, return path to the converted image.
    """
    converted_image_path = '{}-cloud.img'.format(image_path.strip('.img'))
    cmd = ['qemu-img',
           'convert',
           '-f', 'raw',
           '-O', 'qcow2',
           image_path,
           converted_image_path]
    check_call(cmd)
    return converted_image_path


def upload_image_to_glance(nova_image_path, config):
    """
    Upload the nova image to glance, returning the id of the image in glance.
    """
    extra = LOGGING_EXTRA.copy()
    glance = get_glance_client(config)
    image_name = os.path.basename(nova_image_path)
    image = glance.images.create(
        name=image_name,
        data=open(nova_image_path, 'rb'),
        disk_format='qcow2',
        container_format='bare',
    )
    logger.info("Uploaded %s with ID %s.", image_name, image.id, extra=extra)
    return image.id
